from .onclick_representDevice import Onclick_representDevice, RepresentDeviceInterface
from .onclick_animate import Onclick_animate
from .onclick_changeSymbol import Onclick_changeSymbol
from .onclick_copySomething import Onclick_copySomething
from .onclick_delete import Onclick_delete
from .onclick_exportCollection import Onclick_exportCollection
from .onclick_exportToTxt import Onclick_exportToTxt
from .onclick_exportTrace import Onclick_exportTrace
from .onclick_extractPareto import Onclick_extractPareto
from .onclick_measure import Onclick_measure
from .onclick_removeTrace import Onclick_removeTrace
from .onclick_representDevice import Onclick_representDevice
from .onclick_tojson import Onclick_tojson
from .onclickInterface import OnclickInterface
from .representDevice_examples import Represent_opengl, Represent_image, Represent_lines, \
    Represent_brut_attributes, Represent_txt_function
from .animation_examples import Animate_lines, Animate_openGL, Animate_lines_and_text, Animate_openGL_and_text
